/* cfgfile.c -- Parsing a configuration file into data structures.
   $Id: cfgfile.c,v 0.7 1997/12/12 00:13:33 eagle Exp $

   Written by Russ Allbery <rra@stanford.edu>
   Copyright 1997 by the Board of Trustees, Leland Stanford Jr. University

   Functions to parse the configuration file and build a linked list of
   tasks for us to perform.  Should be called once from the main program (or
   could potentially be called multiple times for multiple configuration
   files someday, but that will require some tweaks and freeing internal
   lists each time).  This could probably be written more cleanly in lex or
   yacc, but this is smaller and more portable. */

/***************************************************************************
* Includes and prototypes
***************************************************************************/

#include "cfgfile.h"            /* Prototypes and data structures */
#include "config.h"             /* _ALL_SOURCE */
#include "expand.h"             /* expand() */
#include "utility.h"            /* sysdie(), xmalloc() */

#include <ctype.h>              /* isspace() */
#include <grp.h>                /* getgrnam(), struct group */
#include <pwd.h>                /* getpwnam(), struct passwd */
#include <stdarg.h>             /* va_start(), va_end(), va_list */
#include <stdio.h>              /* FILE, fgets(), fprintf(), vfprintf() */
#include <stdlib.h>             /* realloc(), free(), strtol() */
#include <string.h>             /* strlen(), strtok(), strcpy(), etc. */
#include <sys/types.h>          /* time_t */
#include <time.h>               /* time() */

/* On AIX, we need strings.h to get strcasecmp(). */
#ifdef _ALL_SOURCE
#include <strings.h>
#endif

static void cfgwarn (char *format, ...);
static char *cfgexpand (char *string, char *name);
static char *read_line (FILE *file);
static void parse_variable ();
static char *lookup_variable (char *word);
static void parse_archive (struct class *class);
static struct command *parse_command (char *name);
static void parse_copy (struct class *class);
static gid_t parse_group (char *group);
static char *parse_name (char *name);
static void parse_log (struct class *class);
static uid_t parse_user (char *user);
static struct class *read_class (FILE *config, char *name);


/***************************************************************************
* Local data structures
***************************************************************************/

/* Struct to store variable content mappings from set commands. */
struct variable {
  char *name;
  char *content;
  struct variable *next;
};

/* Stores a linked list of all variables that have been set. */
static struct variable *variables = NULL;

/* The line of the configuration file currently being parsed. */
static char input[1024];

/* The current line number for syntax error messages. */
static int lineno = 0;


/***************************************************************************
* Warning messages
***************************************************************************/

/* Print a warning message for errors encountered during the parse of the
   configuration file.  This is basically warn() from utility.c, but it
   prints out the line number of the configuration file first. */

static void
cfgwarn (char *format, ...)
{
  va_list message;

  fprintf (stderr, "config line %d: ", lineno);
  va_start (message, format);
  vfprintf (stderr, format, message);
  va_end (message);
  fprintf (stderr, "\n");
}


/***************************************************************************
* Percent expansion
***************************************************************************/

/* Check over a string for the invalid %s escape and then pass it along to
   expand() so that % expansion can be done.  The time passed to it will be
   the current time.  Returns NULL on error.  Space for the returned string
   will be malloc()ed and the caller is responsible for freeing. */

static char *
cfgexpand (char *string, char *name)
{
  time_t now;
  char *p;

  /* Grab the current time since expand() needs it. */
  now = time (NULL);

  /* Check for invalid % escapes. */
  for (p = string; *p != '\0'; p++)
    if (*p == '%' && (strchr ("d%mn", *(++p)) == NULL || *p == '\0'))
      {
        cfgwarn ("Invalid %% escape");
        return NULL;
      }

  return expand (string, name, now);
}


/***************************************************************************
* Word parsing
***************************************************************************/

/* Given a string, convert it to a UID.  If it's a number, just convert it
   to a number, otherwise look up the user in the passwd file and determine
   the UID.  We use -1 as an error indicator, which may possibly cause
   problems with the nobody user on some systems. */

static uid_t
parse_user (char *user)
{
  char *end;
  struct passwd *pw;
  uid_t uid;

  uid = strtol (user, &end, 0);
  if (end == user)
    {
      pw = getpwnam (user);
      if (pw != NULL)
        return pw->pw_uid;
      else
        {
          cfgwarn ("Unknown user %s", user);
          return -1;
        }
    }
  else
    return uid;
}


/* Given a string, convert it to a GID.  If it's a number, just convert it
   to a number, otherwise look up the group in the group file and determine
   the GID.  We use -1 as an error indicator, which may possibly cause
   problems with the nobody group on some systems. */

static gid_t
parse_group (char *group)
{
  char *end;
  struct group *gr;
  gid_t gid;

  gid = strtol (group, &end, 0);
  if (end == group)
    {
      gr = getgrnam (group);
      if (gr != NULL)
        return gr->gr_gid;
      else
        {
          cfgwarn ("Unknown group %s", group);
          return -1;
        }
    }
  else
    return gid;
}


/* Uses strtok to pull the next word off of the current configuration file
   line and parse a log name out of it.  Allocates space for and returns the
   name. */

static char *
parse_name (char *name)
{
  char *word;
  
  /* Parse out the name of the log. */
  word = strtok (NULL, " \t\n");
  if (word == NULL)
    {
      cfgwarn ("Syntax error");
      return NULL;
    }
  else
    {
      name = cfgexpand (word, name);
      if (name == NULL)
        {
          cfgwarn ("Unable to expand log name");
          return NULL;
        }
      return name;
    }
}


/***************************************************************************
* Low-level file parsing
***************************************************************************/

/* This function returns the next line of the configuration file, skipping
   over blank lines and comments.  The returned pointer is a pointer into
   static storage, so it can be safely passed around and manipulated without
   worrying about memory leaks or freeing.  Returns NULL on EOF. */

static char *
read_line (FILE *file)
{
  char *start;

  /* Read in the next line, skipping blank lines and comments. */
  do
    {
      lineno++;
      start = fgets (input, 1024, file);
      if (start == NULL) break;

      /* Check for a line that's too long. */
      if (input[strlen (input) - 1] != '\n')
        {
          cfgwarn ("Line too long, ignoring");
          do
            start = fgets (input, 1024, file);
          while (start != NULL && input[strlen (input) - 1] != '\n');
          start = fgets (input, 1024, file);
          if (start == NULL) break;
        }

      /* Skip over initial whitespace. */
      while (isspace (*start)) start++;
    }
  while (*start == '\0' || *start == '#');

  return start;
}


/***************************************************************************
* Setting and looking up variables
***************************************************************************/

/* We support setting variables to arbitrary values in the configuration
   file, which enters them into an internal variable list stored in this
   parser.  This function assumes the remaining tokens that will be returned
   by strtok (NULL, " \t\n") will be:

     <variable> = <value>

   and stores this variable in the internal list, associated with its
   value.  Percent expansion is not done on the variable; that is handled
   later.  No restraints are enforced in <variable> or <value> except that
   they cannot contain embedded whitespace; comparisons will be
   case-insensitive. */

static void
parse_variable ()
{
  struct variable *variable;
  char *word;
  char *name = NULL;
  char *value = NULL;

  /* Parse the rest of the line.  If word is NULL, we've found an error. */
  word = strtok (NULL, " \t\n");
  if (word != NULL)
    {
      name = (char *) xmalloc (strlen (word) + 1);
      strcpy (name, word);
      word = strtok (NULL, " \t\n");
    }
  if (word != NULL && strcmp (word, "=")) word = NULL;
  if (word != NULL) word = strtok (NULL, " \t\n");
  if (word != NULL)
    {
      value = (char *) xmalloc (strlen (word) + 1);
      strcpy (value, word);
    }
  if (word == NULL || strtok (NULL, " \t\n") != NULL)
    {
      cfgwarn ("Syntax error");
      if (name != NULL) free (name);
      if (value != NULL) free (value);
      return;
    }

  /* Put the parsed data into a variable struct and add it to the list. */
  variable = (struct variable *) xmalloc (sizeof (struct variable));
  variable->name = name;
  variable->content = value;
  if (variables == NULL)
    variables = variable;
  else
    {
      variable->next = variables->next;
      variables->next = variable;
    }
}


/* Look up a variable in our list of variables and return a pointer to its
   value if it is there, otherwise return NULL. */

static char *
lookup_variable (char *word)
{
  struct variable *variable;

  for (variable = variables; variable != NULL; variable = variable->next)
    if (!strcasecmp (variable->name, word))
      return (variable->content);
  return NULL;
}


/***************************************************************************
* Reading classes
***************************************************************************/

/* Parse a command (either running a program or sending SIGHUP to a process)
   into a struct command and return it, or returning NULL in event of an
   error.  We assume that strtok (NULL, " \t\n") can be used to obtain the
   next tokens on the line.  hup takes only one argument, but run takes any
   number of them.  Both should undergo % expansion, so they have to be
   checked for %s.  Note that we don't currently support quoting! */

static struct command *
parse_command (char *name)
{
  struct command *command;
  char *word;
  int argc, max;

  /* Allocate space for the new command struct. */
  command = (struct command *) xmalloc (sizeof (struct command));
  command->pidfile = NULL;
  command->argv = NULL;

  /* Check to see if this is a hup or a run. */
  word = strtok (NULL, " \t\n");
  if (word == NULL)
    {
      cfgwarn ("Syntax error, hup or run expected");
      free (command);
      return NULL;
    }
  else if (!strcasecmp (word, "hup"))
    {
      word = strtok (NULL, " \t\n");
      if (word == NULL)
        {
          cfgwarn ("Syntax error, filename expected");
          free (command);
          return NULL;
        }
      command->pidfile = cfgexpand (word, name);
    }
  else if (!strcasecmp (word, "run"))
    {
      /* Parse the rest of the command line into words. */
      command->argv = (char **) xmalloc (3 * sizeof (char *));
      max = 3;
      argc = 0;
      word = strtok (NULL, " \t\n");
      while (word != NULL)
        {
          argc++;
          if (argc >= max)
            {
              max *= 2;
              command->argv = realloc (command->argv, max * sizeof (char *));
              if (command->argv == NULL)
                sysdie ("Unable to allocate memory.");
            }
          command->argv[argc - 1] = cfgexpand (word, name);
          word = strtok (NULL, " \t\n");
        }
      command->argv[argc] = NULL;
    }
  else
    {
      cfgwarn ("Syntax error, hup or run expected");
      free (command);
      return NULL;
    }

  /* Success.  Return the command struct. */
  return command;
}


/* Parse a log declaration line in a log class, assuming that strtok will
   give each word of the line.  The syntax is:

     <name> <location> <owner> <group> <permissions>

   and <location> may be a variable and therefore undergo variable expansion
   unless it begins with '/'.  We allocate a new log struct and return a
   pointer to it. */

static void
parse_log (struct class *class)
{
  char *word, *location, *end;
  struct log *log, *tail;

  /* Allocate space for the new rotate struct and set defaults. */
  log = (struct log *) xmalloc (sizeof (struct log));
  log->name = NULL;
  log->path = NULL;
  log->tmppath = NULL;
  log->archives = NULL;
  log->rotate = 0;
  log->next = NULL;

  /* Parse out the information.  If word == NULL, we've found an error. */
  log->name = parse_name (class->name);
  if (log->name == NULL)
    {
      free (log);
      return;
    }
  word = strtok (NULL, " \t\n");
  if (word != NULL && *word != '/') word = lookup_variable (word);
  if (word != NULL)
    {
      /* Expansion of the location is based on the class name. */
      location = cfgexpand (word, class->name);
      if (location == NULL)
        {
          cfgwarn ("Unable to expand location path");
          free (log->name);
          free (log);
          return;
        }
      log->path = xmalloc (strlen (log->name) + strlen (location) + 2);
      strcpy (log->path, location);
      strcat (log->path, "/");
      strcat (log->path, log->name);
      free (location);
      word = strtok (NULL, " \t\n");
    }
  if (word != NULL)
    {
      log->owner = parse_user (word);
      if (log->owner == (uid_t) -1) word = NULL;
    }
  if (word != NULL) word = strtok (NULL, " \t\n");
  if (word != NULL)
    {
      log->group = parse_group (word);
      if (log->group == (uid_t) -1) word = NULL;
    }
  if (word != NULL) word = strtok (NULL, " \t\n");
  if (word != NULL)
    {
      log->mode = strtol (word, &end, 8);
      if (end == word)
        {
          cfgwarn ("Invalid mode %s", word);
          word = NULL;
        }
    }
  if (word == NULL || strtok (NULL, " \t\n") != NULL)
    {
      cfgwarn ("Syntax error");
      if (log->path != NULL) free (log->path);
      if (log->name != NULL) free (log->name);
      free (log);
      return;
    }

  /* Success and no errors.  We now have a log struct.  Add it in. */
  if (class->logs == NULL)
    class->logs = log;
  else
    {
      /* We want to add to the end of the list to get the ordering of log
         rotation right just in case it matters. */
      tail = class->logs;
      while (tail->next != NULL) tail = tail->next;
      tail->next = log;
    }
  return;
}


/* Parse an archive line in a log class, assuming that strtok will give each
   word of the line.  The syntax is:

     <name> <archive> <count>

   and <archive> may be a variable and therefore undergo variable expansion
   unless it begins with '/'.  <count> may be 1 to not save multiple copies
   or 0 to indicate that that's being taken care of by other means. */

static void
parse_archive (struct class *class)
{
  char *word, *end;
  char *name = NULL;
  char *location = NULL;
  int length;
  struct archive *archive, *tail;
  struct log *log;

  /* Parse out the name of the log. */
  name = parse_name (class->name);
  if (name == NULL) return;

  /* Now, find the log by that name. */
  for (log = class->logs; log != NULL; log = log->next)
    if (!strcmp (log->name, name)) break;
  if (log == NULL)
    {
      cfgwarn ("Unknown log %s", name);
      return;
    }

  /* Create a new archive struct in which to store our information. */
  archive = (struct archive *) xmalloc (sizeof (struct archive));
  archive->path = NULL;
  archive->count = -1;
  archive->next = NULL;

  /* Parse out the information.  If word == NULL, we've found an error. */
  word = strtok (NULL, " \t\n");
  if (word != NULL && *word != '/') word = lookup_variable (word);
  if (word != NULL)
    {
      location = cfgexpand (word, name);
      if (location == NULL)
        {
          cfgwarn ("Unable to expand archive name");
          if (name != NULL) free (name);
          return;
        }
      word = strtok (NULL, " \t\n");
    }
  if (word != NULL)
    {
      archive->count = strtol (word, &end, 0);
      if (end == word)
        {
          cfgwarn ("Invalid count %s", word);
          word = NULL;
        }
    }
  if (word == NULL || strtok (NULL, " \t\n") != NULL)
    {
      cfgwarn ("Syntax error");
      free (archive);
      if (name != NULL) free (name);
      if (location != NULL) free (location);
      return;
    }

  /* All archived logs are compressed, so add the compression extension.
     Currently this is hard-coded to .gz since we want to be using zlib.  If
     we have a count > 0, also add a .0 to the name of the archive. */
  length = strlen (location) + 4;
  if (archive->count > 0) length += 2;
  archive->path = (char *) xmalloc (length);
  strcpy (archive->path, location);
  if (archive->count > 0) strcat (archive->path, ".0");
  strcat (archive->path, ".gz");
  free (location);

  /* Success and no errors.  We now have an archive struct.  Add it in. */
  if (log->archives == NULL)
    log->archives = archive;
  else
    {
      /* We want to add to the end of the list to get the ordering of log
         rotation right just in case it matters. */
      tail = log->archives;
      while (tail->next != NULL) tail = tail->next;
      tail->next = archive;
    }
  log->rotate = 1;
  if (name != NULL) free (name);
  return;
}


/* Parse a copy line in a log class, assuming that strtok will give each
   word of the line.  The syntax is:

     <name> <archive>

   and <archive> may be a variable and therefore undergo variable expansion
   unless it begins with '/'. */

static void
parse_copy (struct class *class)
{
  char *word;
  char *name = NULL;
  char *location = NULL;
  struct archive *archive, *tail;
  struct log *log;

  /* Parse out the name of the log. */
  name = parse_name (class->name);
  if (name == NULL) return;

  /* Now, find the log by that name. */
  for (log = class->logs; log != NULL; log = log->next)
    if (!strcmp (log->name, name)) break;
  if (log == NULL)
    {
      cfgwarn ("Unknown log %s", name);
      return;
    }

  /* Create a new archive struct in which to store our information. */
  archive = (struct archive *) xmalloc (sizeof (struct archive));
  archive->path = NULL;
  archive->count = -1;
  archive->next = NULL;

  /* Parse out the information.  If word == NULL, we've found an error. */
  word = strtok (NULL, " \t\n");
  if (word != NULL && *word != '/') word = lookup_variable (word);
  if (word != NULL)
    {
      location = cfgexpand (word, name);
      if (location == NULL)
        {
          cfgwarn ("Unable to expand archive name");
          if (name != NULL) free (name);
          return;
        }
    }
  if (word == NULL || strtok (NULL, " \t\n") != NULL)
    {
      cfgwarn ("Syntax error");
      free (archive);
      if (name != NULL) free (name);
      if (location != NULL) free (location);
      return;
    }

  /* All archived logs are compressed, so add the compression extension.
     Currently this is hard-coded to .gz since we want to be using zlib.  */
  archive->path = (char *) xmalloc (strlen (location) + 4);
  strcpy (archive->path, location);
  strcat (archive->path, ".gz");
  free (location);

  /* Success and no errors.  We now have an archive struct.  Add it in. */
  if (log->archives == NULL)
    log->archives = archive;
  else
    {
      /* We want to add to the end of the list to get the ordering of log
         rotation right just in case it matters. */
      tail = log->archives;
      while (tail->next != NULL) tail = tail->next;
      tail->next = archive;
    }
  if (name != NULL) free (name);
  return;
}


/* Read in a log class (everything between braces) including all of the
   settings for that particular class.  We populate a struct class and
   return it, or return NULL in the event of some syntax error that makes
   the class invalid.  Memory for the struct class is malloc()ed and the
   caller is responsible for freeing.  We take a pointer to the
   configuration file and the name of the class. */

static struct class *
read_class (FILE *config, char *name)
{
  struct class *class;
  char *line, *directive;

  /* Allocate space for the new class struct. */
  class = (struct class *) xmalloc (sizeof (struct class));
  class->name = (char *) xmalloc (strlen (name) + 1);
  strcpy (class->name, name);
  class->stop = NULL;
  class->restart = NULL;
  class->logs = NULL;
  class->next = NULL;

  /* There can be three starts: stop, restart, and rotate. */
  line = read_line (config);
  while (line != NULL)
    {
      directive = strtok (line, " \t\n");
      if (directive == NULL)
        /* Should be impossible. */
        cfgwarn ("Syntax error");
      else if (!strcasecmp (directive, "stop:"))
        class->stop = parse_command (name);
      else if (!strcasecmp (directive, "restart:"))
        class->restart = parse_command (name);
      else if (!strcasecmp (directive, "log:"))
        parse_log (class);
      else if (!strcasecmp (directive, "archive:"))
        parse_archive (class);
      else if (!strcasecmp (directive, "copy:"))
        parse_copy (class);
      else if (!strcmp (directive, "}"))
        break;
      else
        cfgwarn ("Unknown directive %s", directive);

      line = read_line (config);
    }

  /* If we saw EOF before the end of the class, that's bad. */
  if (line == NULL)
    {
      cfgwarn ("Unexpected end of file");
      free (class);
      return NULL;
    }

  /* Success, return the new class. */
  return class;
}


/***************************************************************************
* Interface
***************************************************************************/

/* The function that should be called by the rest of the program to parse a
   configuration file.  The name of the file is passed in, and a pointer to
   a list of log classes is returned.  Each of those log classes contains
   one or more logs to rotate or truncate.  If no classes without errors are
   found in the configuration file, NULL is returned.  Memory for the linked
   list of log classes is malloc()ed. */

struct class *
read_config (char *filename)
{
  FILE *config;
  char *line, *word, *brace;
  struct class *class;
  struct class *tail = NULL;
  struct class *head = NULL;

  config = fopen (filename, "r");
  if (config == NULL)
    {
      warn ("Could not open configuration file %s", filename);
      return NULL;
    }

  /* Read in and process each line. */
  line = read_line (config);
  while (line != NULL)
    {
      /* If the line starts with "set", it's a variable setting. */
      word = strtok (line, " \t\n");
      if (!strcasecmp (word, "set"))
        parse_variable ();
      else
        {
          /* Otherwise, if it starts with "name" {, it's a block start. */
          brace = strtok (NULL, " \t\n");
          if (brace == NULL || strcmp (brace, "{"))
            cfgwarn ("Syntax error, { expected");
          else if (strtok (NULL, " \t\n"))
            cfgwarn ("End of line expected");
          else
            {
              class = read_class (config, word);

              /* Add the new class to the linked list of classes we have. */
              if (class != NULL)
                {
                  if (head == NULL)
                    head = tail = class;
                  else
                    {
                      tail->next = class;
                      tail = class;
                    }
                }
            }
        }
      line = read_line (config);
    }

  fclose (config);

  /* Return the linked list of log classes. */
  return head;
}
