/* main.c -- Log file rotation and periodic maintenance.
   $Id: main.c,v 0.4 1997/11/01 03:32:02 eagle Exp $

   Written by Russ Allbery <rra@stanford.edu>
   Copyright 1997 by the Board of Trustees, Leland Stanford Jr. University

   This program is designed to rotate, truncate, or otherwise deal with
   system logs on a regular basis, including correctly stopping and
   restarting daemons so that the logs can be moved and compressed safely.
   It takes a configuration file specifying which logs to rotate and what to
   do with them. */

/***************************************************************************
* Includes
***************************************************************************/

#include "cfgfile.h"            /* read_config(), struct class, etc. */
#include "config.h"             /* CONFIG_FILE */
#include "expand.h"             /* expand() */
#include "file.h"               /* rename_log(), copy_log(), etc. */
#include "process.h"            /* run(), hup() */
#include "utility.h"            /* warn() */

#include <stdlib.h>             /* free(), exit() */
#include <unistd.h>             /* unlink() */


/***************************************************************************
* Running commands
***************************************************************************/

/* Run a command (from a struct command), which involves either sending
   SIGHUP to a process from a PID file or running a command with a variable
   number of arguments.  We also free all of the contents of the struct when
   we're done (but not the struct itself). */

static void
run_command (struct command *command)
{
  int n;
  
  if (command != NULL)
    {
      if (command->pidfile != NULL)
        {
          hup (command->pidfile);
          free (command->pidfile);
        }
      else if (command->argv != NULL)
        {
          run (command->argv);
          for (n = 0; command->argv[n] != NULL; n++)
            free (command->argv[n]);
          free (command->argv);
        }
    }
}


/***************************************************************************
* Main routine
***************************************************************************/

/* Read in our configuration file, getting the list of log classes, and then
   step through each one, performing the desired tasks. */

int
main (int argc, char **argv)
{
  int okay;
  struct class *class, *oldclass;
  struct log *log, *oldlog;
  struct archive *archive, *oldarchive;

  /* If we have a command-line argument, assume that's the config file. */
  if (argc > 1)
    class = read_config (argv[1]);
  else
    class = read_config (CONFIG_FILE);

  /* Step through each class and perform the desired operations. */
  while (class != NULL)
    {
      /* Stop the daemon if necessary. */
      if (class->stop != NULL) run_command (class->stop);
      free (class->stop);

      /* Rotate each log and create the new log file. */
      for (log = class->logs; log != NULL; log = log->next)
        {
          if (!log->rotate) continue;
          log->tmppath = rename_log (log->path);
          if (log->tmppath)
            create_log (log->path, log->owner, log->group, log->mode);
        }

      /* Restart the daemon, if necessary. */
      if (class->restart != NULL) run_command (class->restart);
      free (class->restart);

      /* Copy each log to each of its archive locations and compress it. */
      log = class->logs;
      while (log != NULL)
        {
          if ((log->tmppath == NULL && log->rotate)
              || log->archives == NULL)
            warn ("Skipping archiving of %s", log->name);
          else
            {
              archive = log->archives;
              okay = 1;
              while (archive != NULL)
                {
                  if (archive->count > 0)
                    okay &= shuffle_log (archive->path, archive->count);
                  if (archive->count >= 0)
                    okay &= archive_log (log->tmppath, archive->path);
                  else
                    okay &= overwrite_log (log->path, archive->path);
                  free (archive->path);
                  oldarchive = archive;
                  archive = archive->next;
                  free (oldarchive);
                }
              if (log->rotate && okay && unlink (log->tmppath) != 0)
                syswarn ("Cannot delete temporary file %s");
            }
          if (log->name != NULL) free (log->name);
          if (log->path != NULL) free (log->path);
          if (log->tmppath != NULL) free (log->tmppath);
          oldlog = log;
          log = log->next;
          free (oldlog);
        }

      /* Free the memory from this log class and move on to the next one. */
      if (class->name != NULL) free (class->name);
      oldclass = class;
      class = class->next;
      free (oldclass);
    }

  /* Success.  Exit. */
  exit (0);
}
