/* utility.c -- A small set of utility functions I've found useful.
   $Id: utility.c,v 0.1 1997/06/20 02:18:07 eagle Exp $

   Written by Russ Allbery <rra@stanford.edu>
   Copyright 1997 by the Board of Trustees, Leland Stanford Jr. University

   Some general utility functions, such as warn(), die(), and xmalloc(),
   that make safe programming, standard exit conditions, and the like a bit
   easier to do. */

/***************************************************************************
* Includes
***************************************************************************/

#include "utility.h"            /* Prototypes */

#include <errno.h>              /* errno */
#include <stdarg.h>             /* va_list, va_start(), va_end() */
#include <stdio.h>              /* vfprintf(), stderr */
#include <stdlib.h>             /* malloc() */
#include <string.h>             /* strerror() */
#include <sys/types.h>          /* size_t */


/***************************************************************************
* Error messages
***************************************************************************/

/* warn() should be called for any non-fatal error message; it generates an
   error message to stderr.  This is wrapped in a separate function rather
   than just embedding fprintf()s in the code so that if we ever want to do
   something different with error messages, we can. */

void
warn (char *format, ...)
{
  va_list message;

  va_start (message, format);
  vfprintf (stderr, format, message);
  va_end (message);
  fprintf (stderr, "\n");
}


/* Same as warn(), but add a colon, a space, and the results of strerror()
   run on errno.  Should be used for reporting system errors. */

void
syswarn (char *format, ...)
{
  va_list message;

  va_start (message, format);
  vfprintf (stderr, format, message);
  va_end (message);
  fprintf (stderr, ": %s\n", strerror (errno));
}  


/* die() should be called for all fatal error messages.  It's identical to
   warn() but exits with a non-zero exit status after printing out the error
   message.  die() isn't just a wrapper to warn() because of problems with
   varargs. */

void
die (char *format, ...)
{
  va_list message;

  va_start (message, format);
  vfprintf (stderr, format, message);
  va_end (message);
  fprintf (stderr, "\n");
  exit (1);
}


/* Same as die(), but add a colon, a space, and the results of strerror()
   run on errno.  Should be used for reporting system errors. */

void
sysdie (char *format, ...)
{
  va_list message;

  va_start (message, format);
  vfprintf (stderr, format, message);
  va_end (message);
  fprintf (stderr, ": %s\n", strerror (errno));
  exit (1);
}  


/***************************************************************************
* Memory allocation
***************************************************************************/

/* xmalloc() should be used instead of malloc() for all memory allocation in
   the program.  It calls malloc(), checks the return status, and die()s if
   memory allocation failed. */

void *
xmalloc (size_t size)
{
  void *p;

  p = malloc (size);
  if (p == NULL) sysdie ("Unable to allocate memory");
  return p;
}
